/*
 * Copyright (C) 2009 Red Hat, Inc.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "config.h"

#include <sys/types.h>
#include <sys/wait.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <talloc.h>

#include "log.h"
#include "store.h"
#include "store-int.h"
#include "submit.h"
#include "submit-int.h"
#include "subproc.h"

enum cm_external_status {
	STATUS_ISSUED = 0,
	STATUS_WAIT = 1,
	STATUS_REJECTED = 2,
	STATUS_UNREACHABLE = 3,
};

struct cm_submit_state {
	struct cm_submit_state_pvt pvt;
	struct cm_subproc_state *subproc;
};

/* Get a selectable-for-read descriptor we can poll for status changes. */
static int
cm_submit_e_get_fd(struct cm_store_entry *entry, struct cm_submit_state *state)
{
	return cm_subproc_get_fd(entry, state->subproc);
}

/* Try to save a CA-specific identifier for our submitted request.  That is, if
 * it even gave us one. */
static int
cm_submit_e_save_ca_cookie(struct cm_store_entry *entry,
			   struct cm_submit_state *state)
{
	const char *msg;
	talloc_free(entry->cm_ca_cookie);
	entry->cm_ca_cookie = NULL;
	msg = cm_subproc_get_msg(entry, state->subproc, NULL);
	if ((msg != NULL) && (strlen(msg) > 0)) {
		entry->cm_ca_cookie = talloc_strdup(entry, msg);
		if (entry->cm_ca_cookie == NULL) {
			cm_log(1, "Out of memory.\n");
			return -ENOMEM;
		}
		cm_log(1, "Saved cookie.\n");
		return 0;
	} else {
		cm_log(1, "No cookie.\n");
		return -1;
	}
	return -1;
}

/* Check if an attempt to submit the CSR has completed. */
static int
cm_submit_e_ready(struct cm_store_entry *entry, struct cm_submit_state *state)
{
	int status, ready;
	ready = cm_subproc_ready(entry, state->subproc);
	switch (ready) {
	case 0:
		status = cm_subproc_get_exitstatus(entry, state->subproc);
		cm_log(1, "Certificate submission attempt complete.\n");
		if (WIFEXITED(status)) {
			cm_log(1, "Child status = %d.\n", WEXITSTATUS(status));
			return 0;
		} else {
			cm_log(1, "Child exited unexpectedly.\n");
			return 0;
		}
		break;
	default:
		cm_log(1, "Certificate submission still ongoing.\n");
		return -1;
		break;
	}
}

/* Check if the certificate was issued.  If the exit status was 0, it was
 * issued. */
static int
cm_submit_e_issued(struct cm_store_entry *entry, struct cm_submit_state *state)
{
	const char *msg;
	msg = cm_subproc_get_msg(entry, state->subproc, NULL);
	if ((strstr(msg, "-----BEGIN CERTIFICATE-----") != NULL) &&
	    (strstr(msg, "-----END CERTIFICATE-----") != NULL)) {
		talloc_free(entry->cm_cert);
		entry->cm_cert = talloc_strdup(entry, msg);
		cm_log(1, "Certificate issued.\n");
		return 0;
	} else {
		cm_log(1, "No issued certificate read.\n");
		return -1;
	}
}

/* Check if the certificate was issued.  If the exit status was 0, it was
 * issued. */
static int
cm_submit_e_rejected(struct cm_store_entry *entry,
		     struct cm_submit_state *state)
{
	int status;
	status = cm_subproc_get_exitstatus(entry, state->subproc);
	if (WIFEXITED(status) && (WEXITSTATUS(status) == STATUS_REJECTED)) {
		return 0;
	}
	return -1;
}

/* Check if the CA was unreachable.  If the exit status was right, then we
 * never actually talked to the CA. */
static int
cm_submit_e_unreachable(struct cm_store_entry *entry,
			struct cm_submit_state *state)
{
	int status;
	status = cm_subproc_get_exitstatus(entry, state->subproc);
	if (WIFEXITED(status) && (WEXITSTATUS(status) == STATUS_UNREACHABLE)) {
		return 0;
	}
	return -1;
}

/* Done talking to the CA; clean up. */
static void
cm_submit_e_done(struct cm_store_entry *entry, struct cm_submit_state *state)
{
	if (state->subproc != NULL) {
		cm_subproc_done(entry, state->subproc);
	}
	talloc_free(state);
}

/* Attempt to exec the helper. */
struct cm_submit_e_args {
	int error_fd;
	const char *csr, *cookie, *operation;
};

static int
cm_submit_e_main(int fd, struct cm_store_ca *ca, struct cm_store_entry *entry,
		 void *userdata)
{
	struct cm_submit_e_args *args = userdata;
	int i;
	unsigned char u;
	if (entry->cm_template_subject != NULL) {
		setenv("CERTMONGER_REQ_SUBJECT", entry->cm_template_subject, 1);
	}
	if (entry->cm_template_email != NULL) {
		setenv("CERTMONGER_REQ_EMAIL",
		       cm_submit_maybe_joinv(NULL, "\n",
					     entry->cm_template_email),
		       1);
	}
	if (entry->cm_template_hostname != NULL) {
		setenv("CERTMONGER_REQ_HOSTNAME",
		       cm_submit_maybe_joinv(NULL, "\n",
					     entry->cm_template_hostname),
		       1);
	}
	if (entry->cm_template_principal != NULL) {
		setenv("CERTMONGER_REQ_PRINCIPAL",
		       cm_submit_maybe_joinv(NULL, "\n",
					     entry->cm_template_principal),
		       1);
	}
	if ((args->operation != NULL) && (strlen(args->operation) > 0)) {
		setenv("CERTMONGER_OPERATION", args->operation, 1);
	}
	if ((args->csr != NULL) && (strlen(args->csr) > 0)) {
		setenv("CERTMONGER_CSR", args->csr, 1);
	}
	if ((args->cookie != NULL) && (strlen(args->cookie) > 0)) {
		setenv("CERTMONGER_COOKIE", args->cookie, 1);
	}
	cm_log(1, "Running helper \"%s\".\n", ca->cm_ca_external_helper);
	for (i = sysconf(_SC_OPEN_MAX) - 1; i >= 0; i--) {
		if ((i != STDOUT_FILENO) && (i != args->error_fd)) {
			close(i);
		}
	}
	execlp(ca->cm_ca_external_helper, ca->cm_ca_external_helper, NULL);
	u = errno;
	write(args->error_fd, &u, 1);
	return u;
}

/* Start CSR submission using parameters stored in the entry. */
struct cm_submit_state *
cm_submit_e_start_or_resume(struct cm_store_ca *ca,
			    struct cm_store_entry *entry,
			    const char *csr,
			    const char *cookie,
			    const char *operation)
{
	int errorfds[0];
	unsigned char u;
	struct cm_submit_state *state;
	struct cm_submit_e_args args;
	state = talloc_ptrtype(entry, state);
	if (state != NULL) {
		memset(state, 0, sizeof(*state));
		state->pvt.get_fd = cm_submit_e_get_fd;
		state->pvt.save_ca_cookie = cm_submit_e_save_ca_cookie;
		state->pvt.ready = cm_submit_e_ready;
		state->pvt.issued = cm_submit_e_issued;
		state->pvt.rejected = cm_submit_e_rejected;
		state->pvt.unreachable = cm_submit_e_unreachable;
		state->pvt.done = cm_submit_e_done;
		if (pipe(errorfds) != -1) {
			fcntl(errorfds[1], F_SETFD, 1L);
			args.error_fd = errorfds[1];
			args.csr = csr;
			args.cookie = cookie;
			args.operation = operation;
			state->subproc = cm_subproc_start(cm_submit_e_main,
							  ca, entry, &args);
			close(errorfds[1]);
			if (state->subproc == NULL) {
				talloc_free(state);
				state = NULL;
			} else {
				switch (read(errorfds[0], &u, 1)) {
				case 0:
					/* no data = kernel closed-on-exec, so
					 * the helper started */
					break;
				case -1:
					/* huh? */
					cm_log(1, "Unexpected error while "
					       "starting helper \"%s\".",
					       ca->cm_ca_external_helper);
					cm_subproc_done(entry, state->subproc);
					talloc_free(state);
					state = NULL;
					break;
				default:
					cm_log(1, "Error while starting helper "
					       "\"%s\": %s.",
					       ca->cm_ca_external_helper,
					       strerror(u));
					cm_subproc_done(entry, state->subproc);
					talloc_free(state);
					state = NULL;
					break;
				}
			}
			close(errorfds[0]);
		}
	}
	return state;
}

/* Start CSR submission using parameters stored in the entry. */
struct cm_submit_state *
cm_submit_e_start(struct cm_store_ca *ca, struct cm_store_entry *entry)
{
	if ((entry->cm_ca_cookie != NULL) &&
	    (strlen(entry->cm_ca_cookie) > 0)) {
		return cm_submit_e_start_or_resume(ca, entry, entry->cm_csr,
						   entry->cm_ca_cookie, "POLL");
	} else {
		return cm_submit_e_start_or_resume(ca, entry, entry->cm_csr,
						   entry->cm_ca_cookie,
						   "SUBMIT");
	}
}
